// Copyright 2025 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Code generated by generate.go. DO NOT EDIT.

package stdlib

type pkginfo struct {
	name string
	deps string // list of indices of dependencies, as varint-encoded deltas
}

var deps = [...]pkginfo{
	{"archive/tar", "\x03q\x03F=\x01\n\x01$\x01\x01\x02\x05\b\x02\x01\x02\x02\r"},
	{"archive/zip", "\x02\x04g\a\x03\x13\x021=\x01+\x05\x01\x0f\x03\x02\x0f\x04"},
	{"bufio", "\x03q\x86\x01D\x15"},
	{"bytes", "t+[\x03\fH\x02\x02"},
	{"cmp", ""},
	{"compress/bzip2", "\x02\x02\xf6\x01A"},
	{"compress/flate", "\x02r\x03\x83\x01\f\x033\x01\x03"},
	{"compress/gzip", "\x02\x04g\a\x03\x15nU"},
	{"compress/lzw", "\x02r\x03\x83\x01"},
	{"compress/zlib", "\x02\x04g\a\x03\x13\x01o"},
	{"container/heap", "\xbc\x02"},
	{"container/list", ""},
	{"container/ring", ""},
	{"context", "t\\p\x01\x0e"},
	{"crypto", "\x8a\x01pC"},
	{"crypto/aes", "\x10\v\t\x99\x02"},
	{"crypto/cipher", "\x03!\x01\x01 \x12\x1c,Z"},
	{"crypto/des", "\x10\x16 .,\x9d\x01\x03"},
	{"crypto/dsa", "F\x03+\x86\x01\r"},
	{"crypto/ecdh", "\x03\v\r\x10\x04\x17\x03\x0f\x1c\x86\x01"},
	{"crypto/ecdsa", "\x0e\x05\x03\x05\x01\x10\b\v\x06\x01\x03\x0e\x01\x1c\x86\x01\r\x05L\x01"},
	{"crypto/ed25519", "\x0e\x1f\x12\a\x03\b\a\x1cI=C"},
	{"crypto/elliptic", "4@\x86\x01\r9"},
	{"crypto/fips140", "#\x05\x95\x01\x98\x01"},
	{"crypto/hkdf", "0\x15\x01.\x16"},
	{"crypto/hmac", "\x1b\x16\x14\x01\x122"},
	{"crypto/hpke", "\x03\v\x02\x03\x04\x01\f\x01\x05\x1f\x05\a\x01\x01\x1d\x03\x13\x16\x9b\x01\x1c"},
	{"crypto/internal/boring", "\x0e\x02\x0el"},
	{"crypto/internal/boring/bbig", "\x1b\xec\x01N"},
	{"crypto/internal/boring/bcache", "\xc1\x02\x14"},
	{"crypto/internal/boring/sig", ""},
	{"crypto/internal/constanttime", ""},
	{"crypto/internal/cryptotest", "\x03\r\v\b%\x10\x19\x06\x13\x12 \x04\x06\t\x19\x01\x11\x11\x1b\x01\a\x05\b\x03\x05\f"},
	{"crypto/internal/entropy", "K"},
	{"crypto/internal/entropy/v1.0.0", "D0\x95\x018\x14"},
	{"crypto/internal/fips140", "C1\xbf\x01\v\x17"},
	{"crypto/internal/fips140/aes", "\x03 \x03\x02\x14\x05\x01\x01\x05,\x95\x014"},
	{"crypto/internal/fips140/aes/gcm", "#\x01\x02\x02\x02\x12\x05\x01\x06,\x92\x01"},
	{"crypto/internal/fips140/alias", "\xd5\x02"},
	{"crypto/internal/fips140/bigmod", "(\x19\x01\x06,\x95\x01"},
	{"crypto/internal/fips140/check", "#\x0e\a\t\x02\xb7\x01["},
	{"crypto/internal/fips140/check/checktest", "(\x8b\x02\""},
	{"crypto/internal/fips140/drbg", "\x03\x1f\x01\x01\x04\x14\x05\n)\x86\x01\x0f7\x01"},
	{"crypto/internal/fips140/ecdh", "\x03 \x05\x02\n\r3\x86\x01\x0f7"},
	{"crypto/internal/fips140/ecdsa", "\x03 \x04\x01\x02\a\x03\x06:\x16pF"},
	{"crypto/internal/fips140/ed25519", "\x03 \x05\x02\x04\f:\xc9\x01\x03"},
	{"crypto/internal/fips140/edwards25519", "\x1f\t\a\x123\x95\x017"},
	{"crypto/internal/fips140/edwards25519/field", "(\x14\x053\x95\x01"},
	{"crypto/internal/fips140/hkdf", "\x03 \x05\t\a<\x16"},
	{"crypto/internal/fips140/hmac", "\x03 \x15\x01\x01:\x16"},
	{"crypto/internal/fips140/mldsa", "\x03\x1c\x04\x05\x02\x0e\x01\x03\x053\x95\x017"},
	{"crypto/internal/fips140/mlkem", "\x03 \x05\x02\x0f\x03\x053\xcc\x01"},
	{"crypto/internal/fips140/nistec", "\x1f\t\r\f3\x95\x01*\r\x15"},
	{"crypto/internal/fips140/nistec/fiat", "(\x148\x95\x01"},
	{"crypto/internal/fips140/pbkdf2", "\x03 \x05\t\a<\x16"},
	{"crypto/internal/fips140/rsa", "\x03\x1c\x04\x04\x01\x02\x0e\x01\x01\x028\x16pF"},
	{"crypto/internal/fips140/sha256", "\x03 \x1e\x01\x06,\x16\x7f"},
	{"crypto/internal/fips140/sha3", "\x03 \x19\x05\x012\x95\x01L"},
	{"crypto/internal/fips140/sha512", "\x03 \x1e\x01\x06,\x16\x7f"},
	{"crypto/internal/fips140/ssh", "(b"},
	{"crypto/internal/fips140/subtle", "\x1f\a\x1b\xc8\x01"},
	{"crypto/internal/fips140/tls12", "\x03 \x05\t\a\x02:\x16"},
	{"crypto/internal/fips140/tls13", "\x03 \x05\b\b\t3\x16"},
	{"crypto/internal/fips140cache", "\xb3\x02\r'"},
	{"crypto/internal/fips140deps", ""},
	{"crypto/internal/fips140deps/byteorder", "\xa0\x01"},
	{"crypto/internal/fips140deps/cpu", "\xb5\x01\a"},
	{"crypto/internal/fips140deps/godebug", "\xbd\x01"},
	{"crypto/internal/fips140deps/time", "\xcf\x02"},
	{"crypto/internal/fips140hash", "9\x1d4\xcb\x01"},
	{"crypto/internal/fips140only", "\x17\x13\x0e\x01\x01Pp"},
	{"crypto/internal/fips140test", ""},
	{"crypto/internal/impl", "\xbe\x02"},
	{"crypto/internal/rand", "\x1b\x0f s=["},
	{"crypto/internal/randutil", "\xfa\x01\x12"},
	{"crypto/internal/sysrand", "tq! \r\r\x01\x01\r\x06"},
	{"crypto/internal/sysrand/internal/seccomp", "t"},
	{"crypto/md5", "\x0e8.\x16\x16i"},
	{"crypto/mlkem", "\x0e%"},
	{"crypto/mlkem/mlkemtest", "3\x13\b&"},
	{"crypto/pbkdf2", "6\x0f\x01.\x16"},
	{"crypto/rand", "\x1b\x0f\x1c\x03+\x86\x01\rN"},
	{"crypto/rc4", "& .\xc9\x01"},
	{"crypto/rsa", "\x0e\r\x01\v\x10\x0e\x01\x03\b\a\x1c\x03\x133=\f\x01"},
	{"crypto/sha1", "\x0e\r+\x02,\x16\x16\x15T"},
	{"crypto/sha256", "\x0e\r\x1dR"},
	{"crypto/sha3", "\x0e+Q\xcb\x01"},
	{"crypto/sha512", "\x0e\r\x1fP"},
	{"crypto/subtle", "\x1f\x1d\x9f\x01z"},
	{"crypto/tls", "\x03\b\x02\x01\x01\x01\x01\x02\x01\x01\x01\x02\x01\x01\x01\t\x01\x18\x01\x0f\x01\x03\x01\x01\x01\x01\x02\x01\x02\x01\x17\x02\x03\x13\x16\x15\b=\x16\x16\r\b\x01\x01\x01\x02\x01\x0e\x06\x02\x01\x0f"},
	{"crypto/tls/internal/fips140tls", "\x17\xaa\x02"},
	{"crypto/x509", "\x03\v\x01\x01\x01\x01\x01\x01\x01\x017\x06\x01\x01\x02\x05\x0e\x06\x02\x02\x03F\x03:\x01\x02\b\x01\x01\x02\a\x10\x05\x01\x06\a\b\x02\x01\x02\x0f\x02\x01\x01\x02\x03\x01"},
	{"crypto/x509/pkix", "j\x06\a\x90\x01H"},
	{"database/sql", "\x03\nQ\x16\x03\x83\x01\v\a\"\x05\b\x02\x03\x01\x0e\x02\x02\x02"},
	{"database/sql/driver", "\rg\x03\xb7\x01\x0f\x12"},
	{"debug/buildinfo", "\x03^\x02\x01\x01\b\a\x03g\x1a\x02\x01+\x0f "},
	{"debug/dwarf", "\x03j\a\x03\x83\x011\x11\x01\x01"},
	{"debug/elf", "\x03\x06W\r\a\x03g\x1b\x01\f \x17\x01\x17"},
	{"debug/gosym", "\x03j\n$\xa1\x01\x01\x01\x02"},
	{"debug/macho", "\x03\x06W\r\ng\x1c,\x17\x01"},
	{"debug/pe", "\x03\x06W\r\a\x03g\x1c,\x17\x01\x17"},
	{"debug/plan9obj", "m\a\x03g\x1c,"},
	{"embed", "t+B\x19\x01T"},
	{"embed/internal/embedtest", ""},
	{"encoding", ""},
	{"encoding/ascii85", "\xfa\x01C"},
	{"encoding/asn1", "\x03q\x03g(\x01'\r\x02\x01\x11\x03\x01"},
	{"encoding/base32", "\xfa\x01A\x02"},
	{"encoding/base64", "\xa0\x01ZA\x02"},
	{"encoding/binary", "t\x86\x01\f(\r\x05"},
	{"encoding/csv", "\x02\x01q\x03\x83\x01D\x13\x02"},
	{"encoding/gob", "\x02f\x05\a\x03g\x1c\v\x01\x03\x1d\b\x12\x01\x10\x02"},
	{"encoding/hex", "t\x03\x83\x01A\x03"},
	{"encoding/json", "\x03\x01d\x04\b\x03\x83\x01\f(\r\x02\x01\x02\x11\x01\x01\x02"},
	{"encoding/pem", "\x03i\b\x86\x01A\x03"},
	{"encoding/xml", "\x02\x01e\f\x03\x83\x014\x05\n\x01\x02\x11\x02"},
	{"errors", "\xd0\x01\x85\x01"},
	{"expvar", "qLA\b\v\x15\r\b\x02\x03\x01\x12"},
	{"flag", "h\f\x03\x83\x01,\b\x05\b\x02\x01\x11"},
	{"fmt", "tF'\x19\f \b\r\x02\x03\x13"},
	{"go/ast", "\x03\x01s\x0f\x01s\x03)\b\r\x02\x01\x13\x02"},
	{"go/build", "\x02\x01q\x03\x01\x02\x02\b\x02\x01\x17\x1f\x04\x02\b\x1c\x13\x01+\x01\x04\x01\a\b\x02\x01\x13\x02\x02"},
	{"go/build/constraint", "t\xc9\x01\x01\x13\x02"},
	{"go/constant", "w\x10\x7f\x01\x024\x01\x02\x13"},
	{"go/doc", "\x04s\x01\x05\n=61\x10\x02\x01\x13\x02"},
	{"go/doc/comment", "\x03t\xc4\x01\x01\x01\x01\x13\x02"},
	{"go/format", "\x03t\x01\f\x01\x02sD"},
	{"go/importer", "y\a\x01\x02\x04\x01r9"},
	{"go/internal/gccgoimporter", "\x02\x01^\x13\x03\x04\f\x01p\x02,\x01\x05\x11\x01\r\b"},
	{"go/internal/gcimporter", "\x02u\x10\x010\x05\r0,\x15\x03\x02"},
	{"go/internal/scannerhooks", "\x87\x01"},
	{"go/internal/srcimporter", "w\x01\x01\v\x03\x01r,\x01\x05\x12\x02\x15"},
	{"go/parser", "\x03q\x03\x01\x02\b\x04\x01s\x01+\x06\x12"},
	{"go/printer", "w\x01\x02\x03\ns\f \x15\x02\x01\x02\f\x05\x02"},
	{"go/scanner", "\x03t\v\x05s2\x10\x01\x14\x02"},
	{"go/token", "\x04s\x86\x01>\x02\x03\x01\x10\x02"},
	{"go/types", "\x03\x01\x06j\x03\x01\x03\t\x03\x024\x063\x04\x03\t \x06\a\b\x01\x01\x01\x02\x01\x10\x02\x02"},
	{"go/version", "\xc2\x01|"},
	{"hash", "\xfa\x01"},
	{"hash/adler32", "t\x16\x16"},
	{"hash/crc32", "t\x16\x16\x15\x8b\x01\x01\x14"},
	{"hash/crc64", "t\x16\x16\xa0\x01"},
	{"hash/fnv", "t\x16\x16i"},
	{"hash/maphash", "\x8a\x01\x11<~"},
	{"html", "\xbe\x02\x02\x13"},
	{"html/template", "\x03n\x06\x19-=\x01\n!\x05\x01\x02\x03\f\x01\x02\r\x01\x03\x02"},
	{"image", "\x02r\x1fg\x0f4\x03\x01"},
	{"image/color", ""},
	{"image/color/palette", "\x93\x01"},
	{"image/draw", "\x92\x01\x01\x04"},
	{"image/gif", "\x02\x01\x05l\x03\x1b\x01\x01\x01\vZ\x0f"},
	{"image/internal/imageutil", "\x92\x01"},
	{"image/jpeg", "\x02r\x1e\x01\x04c"},
	{"image/png", "\x02\ad\n\x13\x02\x06\x01gC"},
	{"index/suffixarray", "\x03j\a\x86\x01\f+\n\x01"},
	{"internal/abi", "\xbc\x01\x99\x01"},
	{"internal/asan", "\xd5\x02"},
	{"internal/bisect", "\xb3\x02\r\x01"},
	{"internal/buildcfg", "wHg\x06\x02\x05\n\x01"},
	{"internal/bytealg", "\xb5\x01\xa0\x01"},
	{"internal/byteorder", ""},
	{"internal/cfg", ""},
	{"internal/cgrouptest", "w[T\x06\x0f\x02\x01\x04\x01"},
	{"internal/chacha8rand", "\xa0\x01\x15\a\x99\x01"},
	{"internal/copyright", ""},
	{"internal/coverage", ""},
	{"internal/coverage/calloc", ""},
	{"internal/coverage/cfile", "q\x06\x17\x17\x01\x02\x01\x01\x01\x01\x01\x01\x01\"\x02',\x06\a\n\x01\x03\x0e\x06"},
	{"internal/coverage/cformat", "\x04s.\x04Q\v6\x01\x02\x0e"},
	{"internal/coverage/cmerge", "w.a"},
	{"internal/coverage/decodecounter", "m\n.\v\x02H,\x17\x18"},
	{"internal/coverage/decodemeta", "\x02k\n\x17\x17\v\x02H,"},
	{"internal/coverage/encodecounter", "\x02k\n.\f\x01\x02F\v!\x15"},
	{"internal/coverage/encodemeta", "\x02\x01j\n\x13\x04\x17\r\x02F,/"},
	{"internal/coverage/pods", "\x04s.\x81\x01\x06\x05\n\x02\x01"},
	{"internal/coverage/rtcov", "\xd5\x02"},
	{"internal/coverage/slicereader", "m\n\x83\x01["},
	{"internal/coverage/slicewriter", "w\x83\x01"},
	{"internal/coverage/stringtab", "w9\x04F"},
	{"internal/coverage/test", ""},
	{"internal/coverage/uleb128", ""},
	{"internal/cpu", "\xd5\x02"},
	{"internal/dag", "\x04s\xc4\x01\x03"},
	{"internal/diff", "\x03t\xc5\x01\x02"},
	{"internal/exportdata", "\x02\x01q\x03\x02e\x1c,\x01\x05\x11\x01\x02"},
	{"internal/filepathlite", "t+B\x1a@"},
	{"internal/fmtsort", "\x04\xaa\x02\r"},
	{"internal/fuzz", "\x03\nH\x18\x04\x03\x03\x01\f\x036=\f\x03\x1d\x01\x05\x02\x05\n\x01\x02\x01\x01\r\x04\x02"},
	{"internal/goarch", ""},
	{"internal/godebug", "\x9d\x01!\x82\x01\x01\x14"},
	{"internal/godebugs", ""},
	{"internal/goexperiment", ""},
	{"internal/goos", ""},
	{"internal/goroot", "\xa6\x02\x01\x05\x12\x02"},
	{"internal/gover", "\x04"},
	{"internal/goversion", ""},
	{"internal/lazyregexp", "\xa6\x02\v\r\x02"},
	{"internal/lazytemplate", "\xfa\x01,\x18\x02\r"},
	{"internal/msan", "\xd5\x02"},
	{"internal/nettrace", ""},
	{"internal/obscuretestdata", "l\x8e\x01,"},
	{"internal/oserror", "t"},
	{"internal/pkgbits", "\x03R\x18\a\x03\x04\fs\r\x1f\r\n\x01"},
	{"internal/platform", ""},
	{"internal/poll", "tl\x05\x159\r\x01\x01\r\x06"},
	{"internal/profile", "\x03\x04m\x03\x83\x017\n\x01\x01\x01\x11"},
	{"internal/profilerecord", ""},
	{"internal/race", "\x9b\x01\xba\x01"},
	{"internal/reflectlite", "\x9b\x01!;<\""},
	{"internal/runtime/atomic", "\xbc\x01\x99\x01"},
	{"internal/runtime/cgroup", "\x9f\x01=\x04u"},
	{"internal/runtime/exithook", "\xd1\x01\x84\x01"},
	{"internal/runtime/gc", "\xbc\x01"},
	{"internal/runtime/gc/internal/gen", "\nc\n\x18k\x04\v\x1d\b\x10\x02"},
	{"internal/runtime/gc/scan", "\xb5\x01\a\x18\az"},
	{"internal/runtime/maps", "\x9b\x01\x01 \n\t\t\x03z"},
	{"internal/runtime/math", "\xbc\x01"},
	{"internal/runtime/pprof/label", ""},
	{"internal/runtime/startlinetest", ""},
	{"internal/runtime/sys", "\xbc\x01\x04"},
	{"internal/runtime/syscall/linux", "\xbc\x01\x99\x01"},
	{"internal/runtime/wasitest", ""},
	{"internal/saferio", "\xfa\x01["},
	{"internal/singleflight", "\xc0\x02"},
	{"internal/strconv", "\x89\x02L"},
	{"internal/stringslite", "\x9f\x01\xb6\x01"},
	{"internal/sync", "\x9b\x01!\x13r\x14"},
	{"internal/synctest", "\x9b\x01\xba\x01"},
	{"internal/syscall/execenv", "\xc2\x02"},
	{"internal/syscall/unix", "\xb3\x02\x0e\x01\x13"},
	{"internal/sysinfo", "\x02\x01\xb2\x01E,\x18\x02"},
	{"internal/syslist", ""},
	{"internal/testenv", "\x03\ng\x02\x01*\x1b\x0f0+\x01\x05\a\n\x01\x02\x02\x01\f"},
	{"internal/testhash", "\x03\x87\x01p\x118\f"},
	{"internal/testlog", "\xc0\x02\x01\x14"},
	{"internal/testpty", "t\x03\xaf\x01"},
	{"internal/trace", "\x02\x01\x01\x06c\a\x03w\x03\x03\x06\x03\t+\n\x01\x01\x01\x11\x06"},
	{"internal/trace/internal/testgen", "\x03j\nu\x03\x02\x03\x011\v\r\x11"},
	{"internal/trace/internal/tracev1", "\x03\x01i\a\x03}\x06\f5\x01"},
	{"internal/trace/raw", "\x02k\nz\x03\x06C\x01\x13"},
	{"internal/trace/testtrace", "\x02\x01q\x03q\x04\x03\x05\x01\x05,\v\x02\b\x02\x01\x05"},
	{"internal/trace/tracev2", ""},
	{"internal/trace/traceviewer", "\x02d\v\x06\x1a<\x1f\a\a\x04\b\v\x15\x01\x05\a\n\x01\x02\x0f"},
	{"internal/trace/traceviewer/format", ""},
	{"internal/trace/version", "wz\t"},
	{"internal/txtar", "\x03t\xaf\x01\x18"},
	{"internal/types/errors", "\xbd\x02"},
	{"internal/unsafeheader", "\xd5\x02"},
	{"internal/xcoff", "`\r\a\x03g\x1c,\x17\x01"},
	{"internal/zstd", "m\a\x03\x83\x01\x0f"},
	{"io", "t\xcc\x01"},
	{"io/fs", "t+*11\x10\x14\x04"},
	{"io/ioutil", "\xfa\x01\x01+\x15\x03"},
	{"iter", "\xcf\x01d\""},
	{"log", "w\x83\x01\x05'\r\r\x01\x0e"},
	{"log/internal", ""},
	{"log/slog", "\x03\n[\t\x03\x03\x83\x01\x04\x01\x02\x02\x03(\x05\b\x02\x01\x02\x01\x0e\x02\x02\x02"},
	{"log/slog/internal", ""},
	{"log/slog/internal/benchmarks", "\rg\x03\x83\x01\x06\x03:\x12"},
	{"log/slog/internal/buffer", "\xc0\x02"},
	{"log/syslog", "t\x03\x87\x01\x12\x16\x18\x02\x0f"},
	{"maps", "\xfd\x01X"},
	{"math", "\xb5\x01TL"},
	{"math/big", "\x03q\x03)\x15E\f\x03\x020\x02\x01\x02\x15"},
	{"math/big/internal/asmgen", "\x03\x01s\x92\x012\x03"},
	{"math/bits", "\xd5\x02"},
	{"math/cmplx", "\x86\x02\x03"},
	{"math/rand", "\xbd\x01I:\x01\x14"},
	{"math/rand/v2", "t,\x03c\x03L"},
	{"mime", "\x02\x01i\b\x03\x83\x01\v!\x15\x03\x02\x11\x02"},
	{"mime/multipart", "\x02\x01N#\x03F=\v\x01\a\x02\x15\x02\x06\x0f\x02\x01\x17"},
	{"mime/quotedprintable", "\x02\x01t\x83\x01"},
	{"net", "\x04\tg+\x1e\n\x05\x13\x01\x01\x04\x15\x01%\x06\r\b\x05\x01\x01\r\x06\a"},
	{"net/http", "\x02\x01\x03\x01\x04\x02D\b\x13\x01\a\x03F=\x01\x03\a\x01\x03\x02\x02\x01\x02\x06\x02\x01\x01\n\x01\x01\x05\x01\x02\x05\b\x01\x01\x01\x02\x01\x0e\x02\x02\x02\b\x01\x01\x01"},
	{"net/http/cgi", "\x02W\x1b\x03\x83\x01\x04\a\v\x01\x13\x01\x01\x01\x04\x01\x05\x02\b\x02\x01\x11\x0e"},
	{"net/http/cookiejar", "\x04p\x03\x99\x01\x01\b\a\x05\x16\x03\x02\x0f\x04"},
	{"net/http/fcgi", "\x02\x01\n`\a\x03\x83\x01\x16\x01\x01\x14\x18\x02\x0f"},
	{"net/http/httptest", "\x02\x01\nL\x02\x1b\x01\x83\x01\x04\x12\x01\n\t\x02\x17\x01\x02\x0f\x0e"},
	{"net/http/httptrace", "\rLnI\x14\n!"},
	{"net/http/httputil", "\x02\x01\ng\x03\x83\x01\x04\x0f\x03\x01\x05\x02\x01\v\x01\x19\x02\x01\x0e\x0e"},
	{"net/http/internal", "\x02\x01q\x03\x83\x01"},
	{"net/http/internal/ascii", "\xbe\x02\x13"},
	{"net/http/internal/httpcommon", "\rg\x03\x9f\x01\x0e\x01\x17\x01\x01\x02\x1d\x02"},
	{"net/http/internal/testcert", "\xbe\x02"},
	{"net/http/pprof", "\x02\x01\nj\x19-\x02\x0e-\x04\x13\x14\x01\r\x04\x03\x01\x02\x01\x11"},
	{"net/internal/cgotest", ""},
	{"net/internal/socktest", "w\xc9\x01\x02"},
	{"net/mail", "\x02r\x03\x83\x01\x04\x0f\x03\x14\x1a\x02\x0f\x04"},
	{"net/netip", "\x04p+\x01f\x034\x17"},
	{"net/rpc", "\x02m\x05\x03\x10\ni\x04\x12\x01\x1d\r\x03\x02"},
	{"net/rpc/jsonrpc", "q\x03\x03\x83\x01\x16\x11\x1f"},
	{"net/smtp", "\x194\f\x13\b\x03\x83\x01\x16\x14\x1a"},
	{"net/textproto", "\x02\x01q\x03\x83\x01\f\n-\x01\x02\x15"},
	{"net/url", "t\x03Fc\v\x10\x02\x01\x17"},
	{"os", "t+\x01\x19\x03\x10\x14\x01\x03\x01\x05\x10\x018\b\x05\x01\x01\r\x06"},
	{"os/exec", "\x03\ngI'\x01\x15\x01+\x06\a\n\x01\x03\x01\r"},
	{"os/exec/internal/fdtest", "\xc2\x02"},
	{"os/signal", "\r\x99\x02\x15\x05\x02"},
	{"os/user", "\x02\x01q\x03\x83\x01,\r\n\x01\x02"},
	{"path", "t+\xb4\x01"},
	{"path/filepath", "t+\x1aB+\r\b\x03\x04\x11"},
	{"plugin", "t"},
	{"reflect", "t'\x04\x1d\x13\b\x04\x05\x17\x06\t-\n\x03\x11\x02\x02"},
	{"reflect/internal/example1", ""},
	{"reflect/internal/example2", ""},
	{"regexp", "\x03\xf7\x018\t\x02\x01\x02\x11\x02"},
	{"regexp/syntax", "\xbb\x02\x01\x01\x01\x02\x11\x02"},
	{"runtime", "\x9b\x01\x04\x01\x03\f\x06\a\x02\x01\x01\x0e\x03\x01\x01\x01\x02\x01\x01\x01\x02\x01\x04\x01\x10\x18L"},
	{"runtime/coverage", "\xa7\x01S"},
	{"runtime/debug", "wUZ\r\b\x02\x01\x11\x06"},
	{"runtime/metrics", "\xbe\x01H-\""},
	{"runtime/pprof", "\x02\x01\x01\x03\x06`\a\x03$$\x0f\v!\f \r\b\x01\x01\x01\x02\x02\n\x03\x06"},
	{"runtime/race", "\xb9\x02"},
	{"runtime/race/internal/amd64v1", ""},
	{"runtime/trace", "\rg\x03z\t9\b\x05\x01\x0e\x06"},
	{"slices", "\x04\xf9\x01\fL"},
	{"sort", "\xd0\x0192"},
	{"strconv", "t+A\x01r"},
	{"strings", "t'\x04B\x19\x03\f7\x11\x02\x02"},
	{"structs", ""},
	{"sync", "\xcf\x01\x13\x01P\x0e\x14"},
	{"sync/atomic", "\xd5\x02"},
	{"syscall", "t(\x03\x01\x1c\n\x03\x06\r\x04S\b\x05\x01\x14"},
	{"testing", "\x03\ng\x02\x01X\x17\x14\f\x05\x1b\x06\x02\x05\x02\x05\x01\x02\x01\x02\x01\x0e\x02\x04"},
	{"testing/cryptotest", "QOZ\x124\x03\x12"},
	{"testing/fstest", "t\x03\x83\x01\x01\n&\x10\x03\t\b"},
	{"testing/internal/testdeps", "\x02\v\xae\x01/\x10,\x03\x05\x03\x06\a\x02\x0f"},
	{"testing/iotest", "\x03q\x03\x83\x01\x04"},
	{"testing/quick", "v\x01\x8f\x01\x05#\x10\x11"},
	{"testing/slogtest", "\rg\x03\x89\x01.\x05\x10\f"},
	{"testing/synctest", "\xe3\x01`\x12"},
	{"text/scanner", "\x03t\x83\x01,+\x02"},
	{"text/tabwriter", "w\x83\x01Y"},
	{"text/template", "t\x03C@\x01\n \x01\x05\x01\x02\x05\v\x02\x0e\x03\x02"},
	{"text/template/parse", "\x03t\xbc\x01\n\x01\x13\x02"},
	{"time", "t+\x1e$(*\r\x02\x13"},
	{"time/tzdata", "t\xce\x01\x13"},
	{"unicode", ""},
	{"unicode/utf16", ""},
	{"unicode/utf8", ""},
	{"unique", "\x9b\x01!%\x01Q\r\x01\x14\x12"},
	{"unsafe", ""},
	{"vendor/golang.org/x/crypto/chacha20", "\x10]\a\x95\x01*'"},
	{"vendor/golang.org/x/crypto/chacha20poly1305", "\x10\aV\a\xe2\x01\x04\x01\a"},
	{"vendor/golang.org/x/crypto/cryptobyte", "j\n\x03\x90\x01'!\n"},
	{"vendor/golang.org/x/crypto/cryptobyte/asn1", ""},
	{"vendor/golang.org/x/crypto/internal/alias", "\xd5\x02"},
	{"vendor/golang.org/x/crypto/internal/poly1305", "X\x15\x9c\x01"},
	{"vendor/golang.org/x/net/dns/dnsmessage", "t\xc7\x01"},
	{"vendor/golang.org/x/net/http/httpguts", "\x90\x02\x14\x1a\x15\r"},
	{"vendor/golang.org/x/net/http/httpproxy", "t\x03\x99\x01\x10\x05\x01\x18\x15\r"},
	{"vendor/golang.org/x/net/http2/hpack", "\x03q\x03\x83\x01F"},
	{"vendor/golang.org/x/net/idna", "w\x8f\x018\x15\x10\x02\x01"},
	{"vendor/golang.org/x/net/nettest", "\x03j\a\x03\x83\x01\x11\x05\x16\x01\f\n\x01\x02\x02\x01\f"},
	{"vendor/golang.org/x/sys/cpu", "\xa6\x02\r\n\x01\x17"},
	{"vendor/golang.org/x/text/secure/bidirule", "t\xdf\x01\x11\x01"},
	{"vendor/golang.org/x/text/transform", "\x03q\x86\x01Y"},
	{"vendor/golang.org/x/text/unicode/bidi", "\x03\bl\x87\x01>\x17"},
	{"vendor/golang.org/x/text/unicode/norm", "m\n\x83\x01F\x13\x11"},
	{"weak", "\x9b\x01\x98\x01\""},
}

// bootstrap is the list of bootstrap packages extracted from cmd/dist.
var bootstrap = map[string]bool{
	"cmp":                                     true,
	"cmd/asm":                                 true,
	"cmd/asm/internal/arch":                   true,
	"cmd/asm/internal/asm":                    true,
	"cmd/asm/internal/flags":                  true,
	"cmd/asm/internal/lex":                    true,
	"cmd/cgo":                                 true,
	"cmd/compile":                             true,
	"cmd/compile/internal/abi":                true,
	"cmd/compile/internal/abt":                true,
	"cmd/compile/internal/amd64":              true,
	"cmd/compile/internal/arm":                true,
	"cmd/compile/internal/arm64":              true,
	"cmd/compile/internal/base":               true,
	"cmd/compile/internal/bitvec":             true,
	"cmd/compile/internal/bloop":              true,
	"cmd/compile/internal/compare":            true,
	"cmd/compile/internal/coverage":           true,
	"cmd/compile/internal/deadlocals":         true,
	"cmd/compile/internal/devirtualize":       true,
	"cmd/compile/internal/dwarfgen":           true,
	"cmd/compile/internal/escape":             true,
	"cmd/compile/internal/gc":                 true,
	"cmd/compile/internal/importer":           true,
	"cmd/compile/internal/inline":             true,
	"cmd/compile/internal/inline/inlheur":     true,
	"cmd/compile/internal/inline/interleaved": true,
	"cmd/compile/internal/ir":                 true,
	"cmd/compile/internal/liveness":           true,
	"cmd/compile/internal/logopt":             true,
	"cmd/compile/internal/loong64":            true,
	"cmd/compile/internal/loopvar":            true,
	"cmd/compile/internal/mips":               true,
	"cmd/compile/internal/mips64":             true,
	"cmd/compile/internal/noder":              true,
	"cmd/compile/internal/objw":               true,
	"cmd/compile/internal/pgoir":              true,
	"cmd/compile/internal/pkginit":            true,
	"cmd/compile/internal/ppc64":              true,
	"cmd/compile/internal/rangefunc":          true,
	"cmd/compile/internal/reflectdata":        true,
	"cmd/compile/internal/riscv64":            true,
	"cmd/compile/internal/rttype":             true,
	"cmd/compile/internal/s390x":              true,
	"cmd/compile/internal/slice":              true,
	"cmd/compile/internal/ssa":                true,
	"cmd/compile/internal/ssagen":             true,
	"cmd/compile/internal/staticdata":         true,
	"cmd/compile/internal/staticinit":         true,
	"cmd/compile/internal/syntax":             true,
	"cmd/compile/internal/test":               true,
	"cmd/compile/internal/typebits":           true,
	"cmd/compile/internal/typecheck":          true,
	"cmd/compile/internal/types":              true,
	"cmd/compile/internal/types2":             true,
	"cmd/compile/internal/walk":               true,
	"cmd/compile/internal/wasm":               true,
	"cmd/compile/internal/x86":                true,
	"cmd/internal/archive":                    true,
	"cmd/internal/bio":                        true,
	"cmd/internal/codesign":                   true,
	"cmd/internal/dwarf":                      true,
	"cmd/internal/edit":                       true,
	"cmd/internal/gcprog":                     true,
	"cmd/internal/goobj":                      true,
	"cmd/internal/hash":                       true,
	"cmd/internal/macho":                      true,
	"cmd/internal/obj":                        true,
	"cmd/internal/obj/arm":                    true,
	"cmd/internal/obj/arm64":                  true,
	"cmd/internal/obj/loong64":                true,
	"cmd/internal/obj/mips":                   true,
	"cmd/internal/obj/ppc64":                  true,
	"cmd/internal/obj/riscv":                  true,
	"cmd/internal/obj/s390x":                  true,
	"cmd/internal/obj/wasm":                   true,
	"cmd/internal/obj/x86":                    true,
	"cmd/internal/objabi":                     true,
	"cmd/internal/par":                        true,
	"cmd/internal/pgo":                        true,
	"cmd/internal/pkgpath":                    true,
	"cmd/internal/quoted":                     true,
	"cmd/internal/src":                        true,
	"cmd/internal/sys":                        true,
	"cmd/internal/telemetry":                  true,
	"cmd/internal/telemetry/counter":          true,
	"cmd/link":                                true,
	"cmd/link/internal/amd64":                 true,
	"cmd/link/internal/arm":                   true,
	"cmd/link/internal/arm64":                 true,
	"cmd/link/internal/benchmark":             true,
	"cmd/link/internal/dwtest":                true,
	"cmd/link/internal/ld":                    true,
	"cmd/link/internal/loadelf":               true,
	"cmd/link/internal/loader":                true,
	"cmd/link/internal/loadmacho":             true,
	"cmd/link/internal/loadpe":                true,
	"cmd/link/internal/loadxcoff":             true,
	"cmd/link/internal/loong64":               true,
	"cmd/link/internal/mips":                  true,
	"cmd/link/internal/mips64":                true,
	"cmd/link/internal/ppc64":                 true,
	"cmd/link/internal/riscv64":               true,
	"cmd/link/internal/s390x":                 true,
	"cmd/link/internal/sym":                   true,
	"cmd/link/internal/wasm":                  true,
	"cmd/link/internal/x86":                   true,
	"compress/flate":                          true,
	"compress/zlib":                           true,
	"container/heap":                          true,
	"debug/dwarf":                             true,
	"debug/elf":                               true,
	"debug/macho":                             true,
	"debug/pe":                                true,
	"go/build/constraint":                     true,
	"go/constant":                             true,
	"go/version":                              true,
	"internal/abi":                            true,
	"internal/coverage":                       true,
	"cmd/internal/cov/covcmd":                 true,
	"internal/bisect":                         true,
	"internal/buildcfg":                       true,
	"internal/exportdata":                     true,
	"internal/goarch":                         true,
	"internal/godebugs":                       true,
	"internal/goexperiment":                   true,
	"internal/goroot":                         true,
	"internal/gover":                          true,
	"internal/goversion":                      true,
	"internal/lazyregexp":                     true,
	"internal/pkgbits":                        true,
	"internal/platform":                       true,
	"internal/profile":                        true,
	"internal/race":                           true,
	"internal/runtime/gc":                     true,
	"internal/saferio":                        true,
	"internal/syscall/unix":                   true,
	"internal/types/errors":                   true,
	"internal/unsafeheader":                   true,
	"internal/xcoff":                          true,
	"internal/zstd":                           true,
	"math/bits":                               true,
	"sort":                                    true,
}

// BootstrapVersion is the minor version of Go used during toolchain
// bootstrapping. Packages for which [IsBootstrapPackage] must not use
// features of Go newer than this version.
const BootstrapVersion = Version(24) // go1.24.6
